/*********************************************************************
 * Módulo 3. Tutorial de desarrollo con NME. TicTacToe MiniGame
 * Programación Multimedia y Juegos
 * Autor: David Vallejo Fernández    david.vallejo@tegnix.com
 * Autor: Carlos González Morcillo   carlos.gonzalez@tegnix.com
 *
 * You can redistribute and/or modify this file under the terms of the
 * GNU General Public License ad published by the Free Software
 * Foundation, either version 3 of the License, or (at your option)
 * and later version. See <http://www.gnu.org/licenses/>.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  
 *********************************************************************/

package;

import graphics.sprites.Sky;
import ia.Board;
import ia.Board.TPlayer;
import ia.MiniMax;
import util.FpsLabel;
import util.GUIElement;
import networking.NetworkManager;

import nme.text.TextField;
import nme.text.TextFormat;
import nme.text.TextFieldType;

import aze.display.TileLayer;
import aze.display.TileSprite;
import aze.display.TileClip;
import aze.display.SparrowTilesheet;
import nme.Assets;
import nme.display.Sprite;
import nme.events.Event;
import nme.events.MouseEvent;
import nme.Lib;

import nme.geom.Rectangle;

typedef Tab = {          // Representa ficha del tablero gráfico
 background : TileClip,  // Fondo (usado cuando mouse over)
 csprite : TileClip,     // Ficha (X - O) animada
 x : Int, y : Int,       // Posición física del centro (píxeles)
 area : Rectangle        // Rectangle para detectar si fue pinchado
};

enum GState {            // Estado del juego interno
  Menu;                  // Menú inicial (hasta que inicia partida)
  Playing;               // Jugando (alguien pone fichas)
  Records;               // Fin de partida: Game Over + Records
}    

class TicTacToe extends Sprite{
  var _layerBackground:TileLayer;   // Capa para el fondo     
  var _layerGUI:TileLayer;          // Capa para GUI (general)
  var _layerTabs:TileLayer;         // Capa para las fichas
  var _fpsText:FpsLabel;            // Muestra los FPS
  var _previousTime:Int;            // Tiempo desde lanzamiento
  var _newGameTime:Int;            // Tiempo desde nuevo juego
  var _sw:Int; var _sh:Int;         // Ancho y alto de la pantalla

  var _gstate:GState;               // Game State (Estado juego)
  var _board:Board;                 // Clase del tablero (lógica)
  var _vTabs:Array<Tab>;            // Array de fichas (gráfico)
  var _miniMax:MiniMax;             // Implementación de IA MiniMax
  var _currentPlayer:TPlayer;       // Turno? (Jugador actual)

  var _vGUI:Array<GUIElement>;      // Array de elementos GUI
  var _name:TextField;              // Nombre del jugador
  var _difficulty:Int;              // Nivel de dificultad 
  var _initPlay:Int;                // Inicio del juego (tiempo)
  var _endPlay:Int;                 // Fin del juego (tiempo)
  var _timePlay:TextField;          // TextField: muestra tiempo
  var _records:TextField;           // Texto para mostrar records
  var _newRecord:Bool;              // Registro de nuevo récord?
  var _recordsStr:String;           // String con los récords

  // Constructores e Inicialización ===============================
  // ==============================================================
  public function new() { 
    super(); 
    _recordsStr = "";
    haxe.Timer.delay(init, 150);  
  }

  function init():Void {
    loadResources(); 
    createNewGame(); 
    addListeners();
    NetworkManager.getInstance().connectToServer();
    _previousTime = Lib.getTimer();   
  }

  function addListeners():Void {
    Lib.current.stage.addEventListener(MouseEvent.CLICK, onMouseClick);
    Lib.current.stage.addEventListener(Event.ENTER_FRAME, onEnterFrame);
    Lib.current.stage.addEventListener(MouseEvent.MOUSE_MOVE, onMouseMove);
  }

  function loadResources():Void {
    var sheet:String = Assets.getText("assets/BackgroundAtlas.xml");
    var tilesheet:SparrowTilesheet = new SparrowTilesheet
      (Assets.getBitmapData("assets/BackgroundAtlas.png"), sheet);
    _sw = stage.stageWidth; _sh = stage.stageHeight;
    createBackground(tilesheet);
    createGUI(tilesheet);
    var sheet:String = Assets.getText("assets/TabsAtlas.xml");
    var tilesheet:SparrowTilesheet = new SparrowTilesheet
      (Assets.getBitmapData("assets/TabsAtlas.png"), sheet);
    _layerTabs = new TileLayer(tilesheet);  
    _fpsText = new FpsLabel(32, 409, 0x101010, "assets/cafeta.ttf");  
    addChild(_layerTabs.view); 
    addChild(_fpsText);
  }

  function createMiniMax(level:Int=3):Void {
    _currentPlayer = Player_O;
    _miniMax = new MiniMax(level);
  }

  function createNewGame(level:Int=3):Void {
    _difficulty = level;
    _gstate = Menu;
    _newRecord = false;
    if (_vTabs != null) 
      for (e in _vTabs) {
	if (e.csprite != null) e.csprite.visible = false;
	e.background.visible = false;
      }
    createBoard();
    createMiniMax(_difficulty);

    _newGameTime = Lib.getTimer();
  }

  function createBoard():Void {  // Crea tablero (lógico y gráfico)
    _board = new Board();
    _vTabs = new Array<Tab>();

    for (i in 0...9) {
      var xpos = 93 + 131 * i%3;
      var ypos = 88 + 128 * Std.int(i / 3);
      var cTab : Tab = {background:null, csprite:null, 	x:xpos, 
			y:ypos, area:null};
      cTab.background = addClip("select", xpos, ypos); 
      cTab.area = new Rectangle(xpos - cTab.background.width / 2, 
				ypos - cTab.background.height / 2, 
				cTab.background.width, 
				cTab.background.height);
      _vTabs.push(cTab);
    }
  }

  function createBackground(tilesheet:SparrowTilesheet):Void {
    _layerBackground = new TileLayer(tilesheet);  
    addChild(_layerBackground.view); 
    var background = new Sky(_layerBackground, "background");
    _layerBackground.addChild(background);
    background.update(_sw, _sh);
  }

  function createGUI(tilesheet:SparrowTilesheet):Void {
    _layerGUI = new TileLayer(tilesheet);  
    addChild(_layerGUI.view); 
    _vGUI = new Array<GUIElement>();
    _vGUI.push(new GUIElement(_layerGUI,"easy", 590, 150));
    _vGUI.push(new GUIElement(_layerGUI,"medium", 655, 150, true));
    _vGUI.push(new GUIElement(_layerGUI,"impossible", 732, 150));
    _vGUI.push(new GUIElement(_layerGUI,"time", 610, 210, true));
    showGUI("time", false);

    _vGUI.push(new GUIElement(_layerGUI,"won", 610, 220, true));
    _vGUI.push(new GUIElement(_layerGUI,"lost", 610, 220, true));
    _vGUI.push(new GUIElement(_layerGUI,"tied", 610, 220, true));
    showGUI("won", false); 
    showGUI("lost", false); 
    showGUI("tied", false);

    _vGUI.push(new GUIElement(_layerGUI,"newgamesb", 635, 185, true));
    blinkGUI("newgamesb", true);
    _vGUI.push(new GUIElement(_layerGUI,"buttonng", 700, 185, true));
    _vGUI.push(new GUIElement(_layerGUI,"records", 610, 330, true));
    showGUI("records", false);

    _name = new TextField();  // Nombre del jugador ===============
    var font = Assets.getFont ("assets/cafeta.ttf").fontName;
    var tf:TextFormat = new TextFormat(font, 16, 0x303030, true);
    _name.defaultTextFormat = tf;
    _name.htmlText = "Player 1";
    _name.x = 595; _name.y = 103; _name.selectable = true;
    _name.width = 180; _name.height = 30;
    _name.type = TextFieldType.INPUT;
    addChild(_name);

    _timePlay = new TextField(); // Tiempo de juego ===============
    _timePlay.defaultTextFormat = tf;
    _timePlay.htmlText = "";
    _timePlay.x = 550; _timePlay.y = 176; 
    _timePlay.selectable = false;
    _timePlay.width = 80; _timePlay.height = 30;
    addChild(_timePlay);

    _records = new TextField(); // Muestra records ================
    var tf:TextFormat = new TextFormat(font, 16, 0xEEEEEE, true);
    _records.defaultTextFormat = tf;
    _records.htmlText = "";
    _records.x = 465; _records.y = 293; 
    _records.selectable = false;
    _records.width = 320; _records.height = 97;
    addChild(_records);
  }

  // Métodos auxiliares GUI =======================================
  // ==============================================================

  function addClip(id:String,xp:Int,yp:Int,v:Bool=false):TileClip {
    // Devuelve un clip animado para las fichas del tablero
    var clip = new TileClip(_layerGUI, id, 24);
    clip.x = xp; clip.y = yp; clip.visible = v; clip.loop = false;
    _layerTabs.addChild(clip);
    return (clip);
  }

  function getGUIindex(id:String):Int {
    // Dado el identificador, devuelve el índice en _vGUI
    var i = 0; 
    for (e in _vGUI) {if (e._id == id) return i; i++;} return -1; }

  function getGUIidClicked(xp:Float, yp:Float):String {
    // Dadas coordenadas (píxeles), devuelve a qué elemento clickea
    var id : String = null;
    for (e in _vGUI) if (e.isClicked(xp,yp)) id = e._id;
    return id;
  }

  function showGUI(id:String, val:Bool):Void { 
    _vGUI[getGUIindex(id)].visible = val;  }

  function activeGUI(id:String, val:Bool):Void { 
    _vGUI[getGUIindex(id)]._active = val;  }

  function blinkGUI(id:String, val:Bool):Void { 
    _vGUI[getGUIindex(id)]._blink = val;  }

  function getIdTab(xp:Float, yp:Float):Int {
    for (i in 0...9) if (_vTabs[i].area.contains(xp,yp)) return i;
    return -1; }

  function getDifficulty () : String{
    switch (_difficulty) {
    case 1: return "easy";
    case 3: return "medium";
    case 6: return "impossible";
    default: return "medium";
    }
  }

  // Actualización del GUI ========================================
  // ==============================================================

  function updateGUI(time:Float):Void {
    if (_board.isGameOver()) {
      switch (_board.getWinner()) {
      case Empty: showGUI("tied", true);
      case Player_O: showGUI("won", true); 
      case Player_X: showGUI("lost", true);
      }
      showGUI("records", true);
      blinkGUI("newgamesb", true);
      _gstate = Records;
    }
    
    var strTime = Std.string(Std.int((_endPlay-_initPlay)/1000));
    if (_gstate == Playing) _timePlay.htmlText = strTime;
    for (e in _vGUI) { e.update(time); }  
  }

  function activateBackgroundTabs(xp:Float, yp:Float):Void {
    var moves:Array<Int> = _board.getValidMoves();
    for (i in 0...9)
      if ((_vTabs[i].area.contains(xp,yp)) && (Lambda.has(moves, i)))
	_vTabs[i].background.visible = true;
      else _vTabs[i].background.visible = false;
  }

  function updateBackgroundTabs(time:Float):Void {
    for (i in 0...9) if (_vTabs[i].background.visible == true) 
       _vTabs[i].background.alpha = 
	  0.5 + Math.abs(Math.sin(time*0.005))*0.5;  
  }

  function updateTab(index:Int, player:TPlayer):Void {
    if (player == Player_O) name = "circle";  else name = "cross";
    _vTabs[index].csprite = 
      addClip(name, _vTabs[index].x, _vTabs[index].y, true);
  }

  function clickGUI(xp:Float, yp:Float):Void {
    var id = getGUIidClicked(xp, yp); if (id == null) return;
    switch (id) {
    case "easy": 
      _difficulty = 1; 
      activeGUI("easy", true); 
      activeGUI("medium", false);
      activeGUI("impossible", false);
    case "medium":
      _difficulty = 3; 
      activeGUI("easy", false); 
      activeGUI("medium", true);
      activeGUI("impossible", false);
    case "impossible":
      _difficulty = 6; 
      activeGUI("easy", false); 
      activeGUI("medium", false);
      activeGUI("impossible", true);
    case "buttonng":
      blinkGUI("newgamesb", false); 
      showGUI("time", true);
      showGUI("won", false);
      showGUI("lost", false);
      showGUI("tied", false);
      showGUI("records", false);
      createNewGame(_difficulty);
      _gstate = Playing;
      _initPlay = Lib.getTimer();
    }
  }

  // Gestión de Movimientos (Lógica) ==============================
  // ==============================================================

  function changeTurn():Void {
    if (_currentPlayer == Player_O) _currentPlayer = Player_X;
    else _currentPlayer = Player_O;
  }

  function makeMovement(player:TPlayer, move:Int=-9):Void {
    if (move == -9) move = _miniMax.buildTree(_board, player);
    var moves:Array<Int> = _board.getValidMoves();
    if (Lambda.has(moves, move)) {
      _board.makeMove(move, player);
      updateTab(move, player);
      changeTurn();
    }
  }

  // Mostrar Records ==============================================
  // ==============================================================

  function drawRecords():Void {
    var str = "";

    if (_gstate == Records) {
      
      // Obtiene los récords del NetworkManager,
      // el cual los recuperó previamente del servidor.
      var recordsServer = NetworkManager.getInstance().getRecords();

      if (recordsServer != "") {
	var records:Array<String> = recordsServer.split(";");
	records = records.slice(0, records.length - 1);
	var i:Int = 1;
	// Muestra los 5 primeros registros.
	for (r in records) {
	  var aux:Array<String> = r.split(":");
	  str = str + "<b>" + i + ". </b> " + aux[0] 
	    + " (" + aux[2] + " Seg) [" + aux[1] + "]\n";
	  ++i;
	}
      }

    }
    _records.htmlText = str;
  }

  // Update =======================================================
  // ==============================================================
  function onEnterFrame(e:Event):Void {
    var now = Lib.getTimer();
    var eTime = now - _previousTime;
    if (_gstate == Playing) {
      updateBackgroundTabs(now);
      _endPlay = now;
    }
    drawRecords();
    updateGUI(now);
    if ((_currentPlayer == Player_X) && (eTime > 1000)) 
      makeMovement(_currentPlayer);

    _layerBackground.render(); 
    _layerGUI.render();
    _layerTabs.render(); 
    _fpsText.update(); 
  }

  // Events =======================================================
  // ==============================================================
  function onMouseClick(e:MouseEvent):Void {
    clickGUI(e.localX, e.localY);
    if ((_currentPlayer == Player_O) && (_gstate == Playing)) {
      _previousTime = Lib.getTimer();   
      makeMovement(_currentPlayer, getIdTab(e.localX, e.localY));

      // Envío del nuevo récord.
      if (_board.getWinner() == Player_O && _newRecord == false) {
	_newRecord = true;
	NetworkManager.getInstance()
	  .sendRecordToServer(_name.text,
			      getDifficulty(),
			      Std.int((Lib.getTimer() - _newGameTime) / 1000));
      }
    }
  }

  function onMouseMove(event:MouseEvent):Void {
    if (_gstate == Playing) 
      activateBackgroundTabs(event.localX, event.localY);
  }

  // Main =========================================================
  public function main() {
    Lib.current.addChild(new TicTacToe());
  }
}
