/************************************************************************
 * Módulo 3. Tutorial de desarrollo con NME. TicTacToe MiniGame
 * Programación Multimedia y Juegos
 * Autor: David Vallejo Fernández    david.vallejo@tegnix.com
 * Autor: Carlos González Morcillo   carlos.gonzalez@tegnix.com
 *
 * Código fuente adaptado de http://haxe.org/doc/neko/threadserver
 *
 * You can redistribute and/or modify this file under the terms of the
 * GNU General Public License ad published by the Free Software
 * Foundation, either version 3 of the License, or (at your option)
 * and later version. See <http://www.gnu.org/licenses/>.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  
 ***********************************************************************/

package networking;

import sys.net.Host;
import sys.net.Socket;
import haxe.io.Eof;
import cpp.vm.Thread;

class NetworkManager {

  // Variable estática para implementar Singleton.
  private static var _instance:NetworkManager;

  private var _server:String;            // Servidor de récords.
  private var _port:Int;                 // Puerto del servidor de récords.
  private var _client:String;            // Host del cliente.
  private var _clientPort:Int;           // Puerto del cliente.
  private var _socketServer:Socket;      // Socket del server.
  private var _socketClient:Socket;      // Socket de escucha.
  private var _socketIncoming:Socket;    // Socket de peticiones entrantes.
  private var _networkingThread:Thread;  // Hilo para recibir récords.
  private var _recordsStr:String;        // String con los récords.
  
  private function new () {
    _server = "localhost";
    _port = 2048;
    _client = "localhost";
    _clientPort = 1024;
    _socketServer = new Socket();
    _socketClient = new Socket();
    _socketClient.bind(new Host(_client), _clientPort);
    _socketClient.listen(100);
    _recordsStr = "";

    // Hilo para atender peticiones del servidor de récords.
    _networkingThread = Thread.create(networkingThread);
  }

  // Patrón Singleton.
  public static function getInstance () : NetworkManager {
    if (NetworkManager._instance == null)
      NetworkManager._instance = new NetworkManager();
    return NetworkManager._instance;
  }

  // Establece una conexión con el servidor.
  public function connectToServer () {
    try {
      trace("Estableciendo conexión con el servidor...");
      _socketServer.connect(new Host(_server), _port);
      trace("<OK> Conexión establecida con el servidor.");
    }
    catch (exception : Dynamic) {
      trace("Error de conexión con el servidor.");
    }
  }

  // Función con el código del hilo auxiliar
  // para manejar las conexiones entrantes del servidor.
  private function networkingThread() {
    while (true) {
      try {
	// Acepta un nuevo cliente.
	_socketIncoming = _socketClient.accept();
	// Se bloquea a la espera de datos...
	_socketIncoming.waitForRead();
	// Lectura de récords.
	_recordsStr = _socketIncoming.input.readLine();
      }
      // Captura excepción cuando llega a fin de buffer.
      catch (exception : Eof) {
	_socketIncoming.shutdown(true, true);
      }      
    }
  }

  // Solicita la lista de récords al servidor.
  public function getRecordsFromServer () {
    try {
      var msg:String = "request:" + _client + ":" + _clientPort + ".";
      _socketServer.write(msg); 
    }
    catch (unknown : Dynamic) {
      trace("Error de conexión con " + _server);
    }
  }

  // Envía un récord al servidor.
  // Registra nombre, dificultad y tiempo empleado.
  public function sendRecordToServer (name:String, 
				      difficulty:String, 
				      time:Int) {
    var newRecord:String = "";
    newRecord += (name + ":");
    newRecord += (difficulty + ":");
    newRecord += (time + ".");

    try {
      _socketServer.write(newRecord); 
    }
    catch (unknown : Dynamic) {
      trace("Error de conexión con " + _server);
    }

    // Obtiene la lista de récords actualizada.
    getRecordsFromServer();
  }

  public function getRecords () : String {
    return _recordsStr;
  }

  public function shutdown () : Void {
    _socketServer.close();
    _socketClient.close();
    _socketIncoming.close();
  }

}