/************************************************************************
 * Módulo 3. Tutorial de desarrollo con NME. TicTacToe MiniGame
 * Programación Multimedia y Juegos
 * Autor: David Vallejo Fernández    david.vallejo@tegnix.com
 * Autor: Carlos González Morcillo   carlos.gonzalez@tegnix.com
 *
 * Código fuente adaptado de http://haxe.org/doc/neko/threadserver
 *
 * You can redistribute and/or modify this file under the terms of the
 * GNU General Public License ad published by the Free Software
 * Foundation, either version 3 of the License, or (at your option)
 * and later version. See <http://www.gnu.org/licenses/>.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  
 ***********************************************************************/

import neko.Lib;
import neko.net.Socket;
import neko.net.Host;
import neko.net.ThreadServer;
import haxe.io.Bytes;

typedef Client = {
  var id : Int;
}

typedef Message = {
  var str : String;
}

class Server extends ThreadServer<Client, Message> {

  private var _records:Array<String>;

  public function new () {
    super();
    _records = new Array<String>();
  }

  // Crea un nuevo cliente.
  override function clientConnected (s : Socket) : Client {
    var num = Std.random(100);
    Lib.println("Cliente " + num + " desde " + s.peer());
    return { id: num };
  }

  // Gestiona la desconexión de un cliente.
  override function clientDisconnected ( c : Client ) : Void {
    Lib.println("Cliente " + Std.string(c.id) + " desconectado");
  }

  // Lee el mensaje de un cliente concreto.
  override function readClientMessage (c:Client, buf:Bytes, pos:Int, len:Int) {
    // ¿Mensaje completo?
    var complete = false;
    var cpos = pos;
    while (cpos < (pos+len) && !complete) {
      complete = (buf.get(cpos) == 46);
      cpos++;
    }

    // Si no llega un mensaje completo, readClientMessage retorna.
    if( !complete ) 
      return null;

    // Devuelve el mensaje completo.
    var msg:String = buf.readString(pos, cpos-pos);
    return {msg: {str: msg}, bytes: cpos-pos};
  }

  // Gestiona el mensaje del cliente.
  override function clientMessage (c : Client, msg : Message) : Void {
    var msg_rec = msg.str.substr(0, msg.str.length - 1);
    var tokens = msg_rec.split(":");

    // Petición de récords.
    // msg --> request:host:port
    if (tokens[0] == "request") {
      Lib.println(c.id + " solicita el registro de récords.");
      // Envía records actualizados al cliente que envió el suyo.
      sendRecords(tokens[1], Std.parseInt(tokens[2]));
    }
    // Registro de un nuevo récord.   
    else {
      Lib.println(c.id + " registró un nuevo récord: " + msg.str);
      // Añade el récord.
      addRecord(msg.str.substr(0, msg.str.length - 1));
    }
  }
  
  public function addRecord (newRecord:String) : Void {
    _records.push(newRecord);
    _records.sort(sort);
  }

  // Ordena por dificultad, tiempo.
  // Ej.: impossible 12 > impossible 25
  public function sort (a:String, b:String) : Int {
    var aux_a = a.split(":");
    var aux_b = b.split(":");

    if (aux_a[1] == aux_b[1]) {
      if (aux_a[2] < aux_b[2])
	return -1;
      else
	return 1;
    }
    else {
      if (aux_a[1] == "impossible") return -1;
      if (aux_b[1] == "impossible") return 1;
      if (aux_a[1] == "medium")     return -1;
      if (aux_b[1] == "medium")     return 1;
    }

    return 0;
  }

  // Envía los récords al cliente.
  public function sendRecords (host:String, port:Int) : Void {
    var sock = new Socket();
    sock.connect(new Host(host), port);
    
    Lib.println("Enviando récords..."); Sys.sleep(.1);
    sock.write(getUpdatedRecords());                 
    
    sock.close();
  }

  // Devuelve un String con los 5 mejores récords.
  public function getUpdatedRecords (numRecords:Int = 5) : String {
    return buildRecordsStr(_records.slice(0, numRecords));
  }

  private function buildRecordsStr (records:Array<String>) : String {
    var recordsStr:String = "";
    for (record in records)
      recordsStr += (record + ";");
    return recordsStr;
  }

  public static function main() {
      var server:Server = new Server();
      server.run("localhost", 2048);
  }

}